// ============================================================================
// Project:	Deductive Database
// Filename:	strtab.cpp
// Purpose:	String Table
// Last Change:	24.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#ifdef	VER_DUMP
#include "err.h"
#endif
#include "strmem.h"
#include "flexarr.h"

#include "strtab.h"

//=============================================================================
// Private Types:
//=============================================================================

//-----------------------------------------------------------------------------
// strtab_ntype_t: Node types.
//-----------------------------------------------------------------------------

//typedef enum {
//	STRTAB_NTYPE_LEAF,
//	STRTAB_NTYPE_SMALL,
//	STRTAB_NTYPE_LARGE
//} strtab_ntype_t;

typedef char strtab_ntype_t;
#define STRTAB_NTYPE_LEAF	'\0'
#define STRTAB_NTYPE_SMALL	'\1'
#define STRTAB_NTYPE_FULL	'\2'

//-----------------------------------------------------------------------------
// strtab_node_c: Superclass of Nodes.
//-----------------------------------------------------------------------------

class strtab_node_c {
	public:
	strtab_ntype_t NodeType;
};

typedef strtab_node_c *strtab_node_t;

#define STRTAB_NODE_NULL (static_cast<strtab_node_t>(0))

//-----------------------------------------------------------------------------
// strtab_leaf_s: Leaf node.
//-----------------------------------------------------------------------------

class strtab_leaf_c : public strtab_node_c {
	public:

	str_t	Str;
	//int	Length;
	int	Id;
};

typedef strtab_leaf_c *strtab_leaf_t;

#define STRTAB_LEAF_NULL (static_cast<strtab_leaf_t>(0))

//-----------------------------------------------------------------------------
// strtab_small_s: Small branch node.
//-----------------------------------------------------------------------------

#define STRTAB_SMALL_SIZE	4

class strtab_small_c : public strtab_node_c {
	public:

	int	FanOut;  // Number of children (used positions in arrays)
	char	Chars[STRTAB_SMALL_SIZE];
	strtab_node_c *Children[STRTAB_SMALL_SIZE];
};

typedef strtab_small_c *strtab_small_t;

#define STRTAB_SMALL_NULL (static_cast<strtab_small_t>(0))

// For the free list "FreeSmallNodes":
#define STRTAB_SMALL_NEXT	Children[0]

//-----------------------------------------------------------------------------
// strtab_full_s: Full branch node.
//-----------------------------------------------------------------------------

#define STRTAB_FULL_SIZE	256

class strtab_full_c : public strtab_node_c {
	public:

	strtab_node_c *Children[STRTAB_FULL_SIZE];
};

typedef strtab_full_c *strtab_full_t;

#define STRTAB_FULL_NULL (static_cast<strtab_full_t>(0))

//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

strtab_c::strtab_c(str_t tab_name) :
	MemPool(tab_name), StrMem(&MemPool), StrArr(&MemPool)
{
	// Set Magic number:
	CHECK_CODE(Magic = STRTAB_MAGIC);

	// Check parameter:
	CHECK(tab_name != STR_NULL, "strtab_c::constructor: tab_name is NULL");

	// Initialize attributes (which are not nested objects):
	MemPage        = PAGE_NULL;
	FreePtr        = PAGE_U_NULL;
	FreeUnits      = 0;
	FreeSmallNodes = STRTAB_SMALL_NULL;
	NumPages       = 0;
	Root           = STRTAB_NODE_NULL;
	NumLeafNodes   = 0;
	NumSmallNodes  = 0;
	NumFullNodes   = 0;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

strtab_c::~strtab_c()
{
	// Check this object:
	CHECK_VALID("strtab_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Auxiliary Methods for Search Tree:
//=============================================================================

//-----------------------------------------------------------------------------
// make_leaf: Construct new leaf node.
//-----------------------------------------------------------------------------

inline strtab_leaf_t strtab_c::make_leaf(str_t str, int id)
{
	// Get memory:
	strtab_leaf_t leaf;
	if(FreeUnits >= PAGE_UNITS(strtab_leaf_c)) {
		leaf = reinterpret_cast<strtab_leaf_t>(FreePtr);
		FreePtr += PAGE_UNITS(strtab_leaf_c);
		FreeUnits -= PAGE_UNITS(strtab_leaf_c);
	}
	else {
		MemPage = MemPool.alloc_page();
		if(MemPage == PAGE_NULL)
			return STRTAB_LEAF_NULL;
		NumPages++;
		leaf = reinterpret_cast<strtab_leaf_t>(MemPage);
		FreePtr = reinterpret_cast<page_u_t>(MemPage)
				+ PAGE_UNITS(strtab_leaf_c);
		FreeUnits = PAGE_SIZE_UNITS - PAGE_UNITS(strtab_leaf_c);
		CHECK(FreeUnits >= 0, "strtab_c::make_leaf: Page too small");
	}

	// Initialize node:
	leaf->NodeType = STRTAB_NTYPE_LEAF;
	leaf->Str = str;
	leaf->Id  = id;
	NumLeafNodes++;
	return leaf;
}

//-----------------------------------------------------------------------------
// make_small: Construct small node with a single entry.
//-----------------------------------------------------------------------------

inline strtab_small_t strtab_c::make_small(char c)
{
	// Allocate memory for small node:
	strtab_small_t small;
	if(FreeSmallNodes != STRTAB_SMALL_NULL) {
		small = FreeSmallNodes;
		FreeSmallNodes = reinterpret_cast<strtab_small_t>(
					small->STRTAB_SMALL_NEXT);
	}
	else if(FreeUnits >= PAGE_UNITS(strtab_small_c)) {
		small = reinterpret_cast<strtab_small_t>(FreePtr);
		FreePtr += PAGE_UNITS(strtab_small_c);
		FreeUnits -= PAGE_UNITS(strtab_small_c);
	}
	else {
		MemPage = MemPool.alloc_page();
		if(MemPage == PAGE_NULL)
			return STRTAB_SMALL_NULL;
		NumPages++;
		small = reinterpret_cast<strtab_small_t>(MemPage);
		FreePtr = reinterpret_cast<page_u_t>(MemPage)
				+ PAGE_UNITS(strtab_small_c);
		FreeUnits = PAGE_SIZE_UNITS - PAGE_UNITS(strtab_small_c);
		CHECK(FreeUnits >= 0, "strtab_c::make_small: Page too small");
	}

	// Initialize node:
	small->NodeType = STRTAB_NTYPE_SMALL;
	small->FanOut = 1;
	small->Chars[0] = c;
	small->Children[0] = STRTAB_NODE_NULL;

	// Not really necessary, but make defined state:
	for(int i = 1; i < STRTAB_SMALL_SIZE; i++) {
		small->Chars[i] = '\0';
		small->Children[i] = STRTAB_NODE_NULL;
	}

	// For statistics:
	NumSmallNodes++;

	// Return new node:
	return small;
}


//-----------------------------------------------------------------------------
// make_full: Construct full node with data of small node (upgrade).
//-----------------------------------------------------------------------------

inline strtab_full_t strtab_c::make_full(strtab_small_t small)
{
	// Check Parameter:
	CHECK(small != STRTAB_SMALL_NULL,
			"strtab_c::make_full: small node is null");
	CHECK(small->NodeType == STRTAB_NTYPE_SMALL,
			"strtab_c::make_full: small node is not small");
	CHECK(small->FanOut == STRTAB_SMALL_SIZE,
			"strtab_c::make_full: small node not yet full");
	// Get memory:
	strtab_full_t full;
	if(FreeUnits >= PAGE_UNITS(strtab_full_c)) {
		full = reinterpret_cast<strtab_full_t>(FreePtr);
		FreePtr += PAGE_UNITS(strtab_full_c);
		FreeUnits -= PAGE_UNITS(strtab_full_c);
	}
	else {
		// Note: Because full nodes are large, a large part of
		// the current memory page might remain unused.
		// Therefore, we generate small nodes from the rest of
		// the current page (there is anyway a free list for small
		// nodes and they are quite heavily used).
		// It is at least better than doing nothing with this memory.
		// Maybe I should also generate leaf nodes, however that
		// would need another free list.
		while(FreeUnits >= PAGE_UNITS(strtab_small_c)) {
			strtab_small_t node =
				reinterpret_cast<strtab_small_t>(FreePtr);
			FreePtr += PAGE_UNITS(strtab_small_c);
			FreeUnits -= PAGE_UNITS(strtab_small_c);
			node->STRTAB_SMALL_NEXT = FreeSmallNodes;
			FreeSmallNodes = node;
		}

		// Now get a new memory page for the full node we really need:
		MemPage = MemPool.alloc_page();
		if(MemPage == PAGE_NULL)
			return STRTAB_FULL_NULL;
		NumPages++;
		full = reinterpret_cast<strtab_full_t>(MemPage);
		FreePtr = reinterpret_cast<page_u_t>(MemPage)
				+ PAGE_UNITS(strtab_full_c);
		FreeUnits = PAGE_SIZE_UNITS - PAGE_UNITS(strtab_full_c);
		CHECK(FreeUnits >= 0, "strtab_c::make_full: Page too small");
	}

	// Initialize node type:
	full->NodeType = STRTAB_NTYPE_FULL;

	// Initialize array of child node links to NULL:
	strtab_node_t *p = full->Children;
	for(int i = 0; i < 256; i++)
		*p++ = STRTAB_NODE_NULL;

	// Copy existing child node links from given small node:
	for(int i = 0; i < STRTAB_SMALL_SIZE; i++) {
		int c = static_cast<int>(
				static_cast<unsigned char>(small->Chars[i]));
		CHECK(c >= 0, "strtab_c::make_full: index negative");
		CHECK(c <= 255, "strtab_c::make_full: index too large");
		full->Children[c] = small->Children[i];
	}

	// Free memory for small node:
	small->STRTAB_SMALL_NEXT = FreeSmallNodes;
	FreeSmallNodes = small;

	// For statistics:
	NumSmallNodes--;
	NumFullNodes++;

	// Return new full node:
	return full;
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// put: Insert a string into this string memory pool, return ID.
//-----------------------------------------------------------------------------


int strtab_c::put(str_t start, str_t end) {
	// Check this object:
	CHECK_VALID("strtab_c::put");

	// Check parameters:
	CHECK(start != STR_NULL, "strtab_c::put: start is NULL");
	CHECK(end != STR_NULL, "strtab_c::put: end is NULL");
	CHECK(start <= end, "strtab_c::put: start not before end");

	// Start search with root node:
	strtab_node_t node = Root;
	strtab_node_t *last_node_ptr = &Root;
	for(str_t str_p = start; node != STRTAB_NODE_NULL; str_p++) {
		// It is guaranteed that if str_p == end,
		// *str_p will not be accessed,
		// and node will be set to STRTAB_NODE_NULL.
		// But because this situation corresponds to the '\0' byte
		// normally marking the end of the string, it is still
		// to execute this loop one last time.

		switch(node->NodeType) {
		case STRTAB_NTYPE_LEAF:
			{
			strtab_leaf_t leaf =
					reinterpret_cast<strtab_leaf_t>(node);
			int i = str_p - start;
			str_t s = str_p;
			while(s < end && *s == leaf->Str[i]) {
				s++;
				i++;
			}
			if(s == end && leaf->Str[i] == '\0') {
				// Found:
				return leaf->Id;
			}
			// Leaf does not match search string.
			// Must insert path consisting of small nodes
			// until the position where they differ:
			i = str_p - start;
			strtab_small_t small;
			while(str_p < end && *str_p == leaf->Str[i]) {
				small = make_small(*str_p);
				if(small == STRTAB_SMALL_NULL) {
					// TODO: Restore old state
					return -1;
				}
				*last_node_ptr = small;
				last_node_ptr = small->Children;
						// i.e. Address of Children[0]
				str_p++;
				i++;
			}
			small = make_small(leaf->Str[i]);
			small->Children[0] = leaf;
			small->FanOut = 2;
			small->Chars[1] = (str_p == end) ? '\0' : *str_p;
			*last_node_ptr = small;
			last_node_ptr = &(small->Children[1]);
			node = STRTAB_NODE_NULL; // Rest done below loop
			}
			break;

		case STRTAB_NTYPE_SMALL:
			{
			strtab_small_t small =
					reinterpret_cast<strtab_small_t>(node);
			char search_char = (str_p == end) ? '\0' : *str_p;


			// Check whether there is a child node for this char:
			int i;
			for(i = 0; i < small->FanOut; i++) {
				if(search_char == small->Chars[i]) {
					last_node_ptr = &(small->Children[i]);
					node = small->Children[i];
					break;
				}
			}
			if(i < small->FanOut) {
				// Found character.
				// If str_p == end, we must make sure
				// that there is no further round through
				// the loop.
				if(search_char == '\0') {
					CHECK(node->NodeType ==
						STRTAB_NTYPE_LEAF,
					  "strtab_c::put: (Small) no leaf");
					return reinterpret_cast<strtab_leaf_t>
							(node) -> Id;
				}
				break;
			}

			// Next character of string not in node, must insert:
			if(small->FanOut < STRTAB_SMALL_SIZE) {
				small->Chars[small->FanOut] = search_char;
				last_node_ptr =
					&(small->Children[small->FanOut]);
				small->FanOut++;
				node = STRTAB_NODE_NULL; // Rest below loop
				break;
			}

			// Small node too small, must upgrade to full node:
			strtab_full_t full = make_full(small);
			*last_node_ptr = full;
			if(full == STRTAB_FULL_NULL)
				return -1;
			i = static_cast<int>(
				static_cast<unsigned char>(search_char));
			CHECK(i >= 0, "strtab_c::put: index negative");
			CHECK(i <= 255, "strtab_c::put: index too large");
			last_node_ptr = &(full->Children[i]);
			node = STRTAB_NODE_NULL; // Rest below loop
			}
			break;

		case STRTAB_NTYPE_FULL:
			{
			strtab_full_t full =
					reinterpret_cast<strtab_full_t>(node);
			if(full == STRTAB_FULL_NULL)
				return -1;
			if(str_p != end) {
				int i = static_cast<int>(
					static_cast<unsigned char>(*str_p));
				CHECK(i > 0,
					"strtab_c::put: (Full) index <= 0");
				CHECK(i <= 255,
					"strtab_c::put: (Full) index >255");
				last_node_ptr = &(full->Children[i]);
				node = full->Children[i];
			}
			else {
				// If str_p == end, we must make sure that
				// there is no further round through the loop.
				node = full->Children[0];
				if(node == STRTAB_NODE_NULL)
					last_node_ptr = &(full->Children[0]);
				else {
					// '\0' -> can only be a leaf:
					CHECK(node->NodeType ==
							STRTAB_NTYPE_LEAF,
					  "strtab_c::put: (Full) no leaf");
					return reinterpret_cast<strtab_leaf_t>
							(node) -> Id;
				}
			}
			}
			break;

		default:
			CHECK_IMPOSSIBLE("strtab_c::put: Invalid NodeType");
			return -1;
		}
	}

	// String not found.
	// last_node_ptr points to link, where new leaf must be inserted.

	// First make a copy of the string to our own memory:
	str_t saved_str = StrMem.insert(start, end);
	if(saved_str == STR_NULL) {
		*last_node_ptr = STRTAB_NODE_NULL;
		return -1;
	}

	// Enter address of copied string into flexible array, return index:
	int id = StrArr.append(saved_str);
	if(id < 0) {
		*last_node_ptr = STRTAB_NODE_NULL;
		return -1;
	}

	// Now create the leaf node with this information:
	*last_node_ptr = make_leaf(saved_str, id);
	if(*last_node_ptr != STRTAB_NODE_NULL)
		return id;
	else
		return -1;
}



//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	strtab_c::check() const
{
	// Check magic number:
	if(Magic != STRTAB_MAGIC)
		return "wrong magic number";

	// Numbers cannot be negative:
	if(NumPages < 0)
		CHECK_ERR("NumPages is negative");
	if(NumLeafNodes < 0)
		CHECK_ERR("NumLeafNodes is negative");
	if(NumSmallNodes < 0)
		CHECK_ERR("NumSmallNodes is negative");
	if(NumFullNodes < 0)
		CHECK_ERR("NumFullNodes is negative");

	// If there is a memory page, NumPages cannot be 0, and vice versa:
	if(MemPage != PAGE_NULL && NumPages == 0)
		CHECK_ERR("A page has been allocated, but NumPages is 0");
	if(MemPage == PAGE_NULL && NumPages > 0)
		CHECK_ERR("NumPages > 0, but current memory page is null");

	// If there is a node, there must be a memory page:
	if(NumLeafNodes+NumSmallNodes+NumFullNodes > 0 && MemPage == PAGE_NULL)
		CHECK_ERR("There are nodes, but no memory page was allocated");

	// FreePtr must point into MemPage:
	if(MemPage != PAGE_NULL &&
		FreePtr < reinterpret_cast<page_u_t>(MemPage))
		CHECK_ERR("FreePtr outside bounds (too small)");
	if(MemPage != PAGE_NULL &&
		FreePtr > reinterpret_cast<page_u_t>(MemPage) +PAGE_SIZE_UNITS)
		CHECK_ERR("FreePtr outside bounds (too large)");

	// FreeBytes can be computed from FreePtr:
	if(MemPage == PAGE_NULL && FreeUnits != 0)
		CHECK_ERR("MemPage is null, but FreeUnits != 0");
	if(MemPage != PAGE_NULL &&
		FreeUnits != PAGE_SIZE_UNITS -
				(FreePtr-reinterpret_cast<page_u_t>(MemPage)))
		CHECK_ERR(
		    "FreeUnits has wrong value (inconsistent with FreePtr)");

	// Check sub-objects:
	CHECK_SUBOBJ(MemPool, "strtab_c::check");
	CHECK_SUBOBJ(StrMem, "strtab_c::check");
	CHECK_SUBOBJ(StrArr, "strtab_c::check");

	// Ok:
	return 0;
}

#endif

//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP

void strtab_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "String Table";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:        '");
	err_c::dump_str(MemPool.name());
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("NumStrings:   ");
	err_c::dump_int(StrMem.num_strings());
	err_c::dump_str(" [Should be equal to flexArray length: ");
	err_c::dump_int(StrArr.length());
	err_c::dump_str(" ]");
	err_c::dump_nl();

	err_c::dump_str("Total Memory: ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("String Memory: ");
	pages = StrMem.num_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();

	err_c::dump_str("MemPage:     ");
	err_c::dump_ptr(MemPage);
	err_c::dump_str(" [Current Memory Page for Allocating Nodes]");
	err_c::dump_nl();

	err_c::dump_str("Page Size:   ");
	err_c::dump_int(VER_PAGESIZE);
	err_c::dump_str(" Bytes [");
	err_c::dump_int(PAGE_SIZE_UNITS);
	err_c::dump_str(" units of ");
	err_c::dump_int(PAGE_UNIT_BYTES);
	err_c::dump_str(" Bytes each]");
	err_c::dump_nl();

	err_c::dump_str("FreePtr:     ");
	err_c::dump_ptr(FreePtr);
	err_c::dump_str(" [corresponds to Index: ");
	int ptr_i = static_cast<int>(FreePtr -
					reinterpret_cast<page_u_t>(MemPage));
	err_c::dump_int(ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_str("FreeUnits:   ");
	err_c::dump_int(FreeUnits);
	err_c::dump_str(" [Sum with FreePtr Index: ");
	err_c::dump_int(FreeUnits + ptr_i);
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_str("Object Size: ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strtab_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// String table:
	err_c::dump_str("Strings:");
	err_c::dump_nl();
	for(int i = 0; i < StrArr.length(); i++) {
		err_c::dump_str("    ");
		err_c::dump_int(i);
		err_c::dump_str(". '");
		err_c::dump_str(StrArr.get(i));
		err_c::dump_str("'");
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Search tree:
	err_c::dump_str("Search tree (index):");
	err_c::dump_nl();
	if(Root == STRTAB_NODE_NULL) {
		err_c::dump_str("Root pointer is NULL");
		err_c::dump_nl();
	}
	else {
		(void) dump_node(1, -1, Root, 0, 1);
	}

	// Print node numbers:
	err_c::dump_nl();
	err_c::dump_str("Tree size:");
	err_c::dump_nl();
	err_c::dump_str("    Leaf nodes:  ");
	err_c::dump_int(NumLeafNodes);
	err_c::dump_nl();
	err_c::dump_str("    Small nodes: ");
	err_c::dump_int(NumSmallNodes);
	err_c::dump_nl();
	err_c::dump_str("    Full nodes:  ");
	err_c::dump_int(NumFullNodes);
	err_c::dump_nl();

	// Print node sizes:
	err_c::dump_nl();
	err_c::dump_str("Node sizes:");
	err_c::dump_nl();
	err_c::dump_str("    Base class:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strtab_node_c)));
	err_c::dump_nl();
	err_c::dump_str("    Leaf nodes:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strtab_leaf_c)));
	err_c::dump_nl();
	err_c::dump_str("    Small nodes: ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strtab_small_c)));
	err_c::dump_nl();
	err_c::dump_str("    Full nodes:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(strtab_full_c)));
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}

#endif


//=============================================================================
// Auxiliary Method for Data Structure Dump:
//=============================================================================

//-----------------------------------------------------------------------------
// dump_node: Recursive method to show search tree data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP

int strtab_c::dump_node(int depth, int ch, strtab_node_t node,
					int parent_no, int node_no) const
{
	// Print node number and parent number:
	err_c::dump_str(str_fill(node_no, 4));
	err_c::dump_int(node_no);
	err_c::dump_str(". [^");
	err_c::dump_int(parent_no);
	err_c::dump_str("] ");
	err_c::dump_str(str_fill(parent_no, 4));
	int next_node_no = node_no + 1;

	// Indent node corresponding to depth:
	for(int i = 0; i < depth; i++)
		err_c::dump_str("  ");

	// Print character which leads to this node:
	if(ch == -1)
		err_c::dump_str("ROOT: ");
	else if(ch == 0)
		err_c::dump_str("'\\0': ");
	else if(ch == ' ')
		err_c::dump_str("' ': ");
	else if(ch == '\n')
		err_c::dump_str("'\\n': ");
	else if(ch == '\r')
		err_c::dump_str("'\\r': ");
	else if(ch == '\t')
		err_c::dump_str("'\\t': ");
	else if(ch < 32) {
		err_c::dump_str("'^");
		err_c::dump_char(static_cast<char>(ch+64));
		err_c::dump_str("': ");
	}
	else if(ch == 127)
		err_c::dump_str("'DEL':");
	else {
		err_c::dump_str(" ");
		err_c::dump_char(static_cast<char>(ch));
		err_c::dump_str(":   ");
	}
	
	// Print node, recurse over children:
	strtab_leaf_t leaf;
	strtab_small_t small;
	strtab_full_t full;
	switch(node->NodeType) {
		case STRTAB_NTYPE_LEAF:
			leaf = reinterpret_cast<strtab_leaf_t>(node);
			err_c::dump_str("LEAF: '");
			err_c::dump_str(leaf->Str);
			err_c::dump_str("', Id: ");
			err_c::dump_int(leaf->Id);
			err_c::dump_nl();
			break;

		case STRTAB_NTYPE_SMALL:
			small = reinterpret_cast<strtab_small_t>(node);
			err_c::dump_str("SMALL (FanOut: ");
			err_c::dump_int(small->FanOut);
			err_c::dump_str(")");
			err_c::dump_nl();
			for(int i = 0; i < small->FanOut; i++) {
				next_node_no =
					dump_node(depth + 1, small->Chars[i],
							small->Children[i],
							node_no, next_node_no);
			}
			break;


		case STRTAB_NTYPE_FULL:
			full = reinterpret_cast<strtab_full_t>(node);
			err_c::dump_str("FULL:");
			err_c::dump_nl();
			for(int i = 0; i < 256; i++) {
				if(full->Children[i]) {
					next_node_no =
						dump_node(depth + 1, i,
							full->Children[i],
							node_no, next_node_no);
				}
			}
			break;

		default:
			CHECK_IMPOSSIBLE(
				"strtab_c::dump_node: Invalid NodeType");
			err_c::dump_str("*** INVALID NODE TYPE ***");
			err_c::dump_nl();
	}

	// Return next node number (still unused):
	return next_node_no;
}

#endif

